import os
import re
import types

from amilib.amiweb.amigration import *
from amilib.amiweb import amiweb
from amilib import json

from skeletonz.model import CMSModel

from skeletonz.modules.plugin import GenericPlugin
from amilib.template import render
from skeletonz.mylib.amicache import AmiCache
from skeletonz.server import getConfig, plugin_configurator, getFormatManager
from skeletonz.mylib import html_helpers, converters
from skeletonz.modules.template import PluginTemplate

from skeletonz.user_plugins import TextOption

PLUGINS_FOR_EXPORT = ['PageTabs']

class PageTabs(GenericPlugin):
    NAME = "Page tabs"
    DESCRIPTION = "Gives ability to have page tab navigation."
    SYNTAX = [
        {'handler': 'pagetabs',
         'required_arguments': {'ident': 'The identification'},
         'optional_arguments': {}
        }
      ]

    def __init__(self):
        self.mi_control = AMigrationControl("PageTab",\
            [PageTab_Initial, PageTab_AddHostPage, PageTab_AddHostPageName], plugin_configurator)
        getFormatManager().registerSLPlugin('pagetabs', self.handle)

    def _addStyle(self, template):
        style = """
        #navigation {
          display: none;
        }
        """
        def fn_eval(page_obj):
            if page_obj.content:
                return page_obj.content.find("[pagetabs") != -1
            return True
        template.getHeader().appendStyleData(style, True, fn_eval)

    def addToSiteTemplate(self, template, on_init):
        self._addStyle(template)

    def addToSiteEditTemplate(self, template, on_init):
        ns = {}
        script = render("skeletonz/plugins/pagetabs/view/script.tmpl", ns)
        template.getHeader().appendStaticScriptData(script)
        self._addStyle(template)

    def handle(self, args, edit_mode, page):
        if args.get('pagetabs', '') == '':
            return None

        ident = args.get('pagetabs')
        as_list = args.get('as_list', False)

        t = PageTabsRender.render(ident, edit_mode, as_list)
        return True, t

    def addToController(self, rc):
        path = "%s/skeletonz/plugins/pagetabs/static" % os.getcwd()
        rc.addStaticPath("/static_plugin/pagetabs/", path)
        rc.root_obj.plugin.pagetabs = PageTabsController()

    def createStructure(self):
        self.mi_control.upgradeToLatest()

    def dropStructure(self):
        self.mi_control.downgradeTo(Not_Created)


class PageTabsDeco:

    def toJSON(self):
        d = {'id': self.id,
            'content': self.name}
        return json.write(d)


class PageTabsModel:
    """
    Handles the page tab model queries
    """
    def getTabByIdent(self, ident):
        return amiweb.db().select("plugin_tabs", ident=ident, as_one=True, obj_deco=PageTabsDeco)

    def getTabById(self, id):
        return amiweb.db().select("plugin_tabs", id=id, as_one=True, obj_deco=PageTabsDeco)

    def getPagesByIdent(self, ident):
        page_tab = self.getTabByIdent(ident)
        if not page_tab:
            return []
        result = []
        remove_list = [] #The items we didn't find
        for p_id in page_tab.pages.split(","):
            if p_id != "":
                p = CMSModel.Pages.getPageById(p_id)
                if p:
                    result.append(p)
                else:
                    remove_list.append(p_id)
        self.deleteById(ident, remove_list)
        return result

    def add(self, ident, page_id):
        #Check if the ident is inserted
        page_tab = self.getTabByIdent(ident)
        if not page_tab:
            #Create the initial
            p = amiweb.db().insert("plugin_tabs",
              ident=ident,
              pages=",",
              host_page_id=amiweb.session()['current_page_id']
            )
        else:
            #Append to the list
            if page_tab.pages.find("%s," % page_id) == -1:
                new_pages = "%s%s," % (page_tab.pages, page_id)
                amiweb.db().update("plugin_tabs", id=page_tab.id, pages=new_pages)

        return CMSModel.Pages.getPageById(page_id)

    def updateHostPage(self, p_tab_id, new_host_pid):
        amiweb.db().update("plugin_tabs", id=p_tab_id, host_page_id=new_host_pid)

    def updateHostPageName(self, p_tab_id, new_val):
        amiweb.db().update("plugin_tabs", id=p_tab_id, host_page_name=new_val)

    def deleteById(self, ident, page_ids):
        if type(page_ids) == type(""):
            page_ids = [page_ids]
        page_tab = self.getTabByIdent(ident)
        p = page_tab.pages
        for pid in page_ids:
            p = p.replace(pid, "", 1).replace(",,", ",")
        amiweb.db().update("plugin_tabs", id=page_tab.id, pages=p)

    def swapPageOrder(self, ident, p1, p2):
        old_p = amiweb.db().select("plugin_tabs", ident=ident, as_one=True).pages
        p = old_p.replace(",%s," % p1, ",HOLDER,")
        p = p.replace(",%s," % p2, ",%s," % p1)
        p = p.replace(",HOLDER,", ",%s," % p2)
        amiweb.db().update("plugin_tabs", "ident='%s'" % ident, pages=p)


class PageTabsController:
    """
    Exposes a number of functions on /plugin/pagetabs/
    """
    def __init__(self):
        template = PluginTemplate("pagetabs")
        self.header = template.getHeader()
        self.footer = template.getFooter()

    @amiweb.expose
    def viewManage(self, ident, as_list):
        ns = {'site': self,
              'ident': ident,
              'as_list': as_list,
              'host_page_name': PageTabsModel.getTabByIdent(ident).host_page_name,
              'pages': PageTabsModel.getPagesByIdent(ident)}
        return render("skeletonz/plugins/pagetabs/view/manage.tmpl", ns)

    @amiweb.expose
    def getHTMLData(self, ident, as_list):
        return PageTabsRender.render(ident, True, as_list)

    @amiweb.expose
    def delete(self, ident, id):
        PageTabsModel.deleteById(ident, id)
        CMSModel.Pages.deletePageById(id)
        self._expireCache(ident)
        return "ok"

    @amiweb.expose
    def add(self, ident, name):
        parent = amiweb.session()['current_page_id']
        p = CMSModel.Pages.insertPageFromWiki(parent, name)
        cnt = '[pagetabs=%s]\n\n\n' % ident
        amiweb.db().update("page", id=p.id, content=cnt)
        self._expireCache(ident)
        return PageTabsModel.add(ident, p.id).toJSON()

    @amiweb.expose
    def update(self, id, name, ident=None):
        try:
            id = int(id)
            CMSModel.Pages.setPageName(id, name)
            CMSModel.Pages.setPageTitle(id, name)
            return CMSModel.Pages.getPageById(id).toJSON()
        except ValueError:
            #id is the ident in this...
            p_tab = PageTabsModel.getTabByIdent(id)
            PageTabsModel.updateHostPageName(p_tab.id, name)
            return self._getHostPageJSON(id)

    @amiweb.expose
    def setNewOrder(self, item1, item2, ident):
        item1 = json.read(item1)
        item2 = json.read(item2)

        try:
            item1_json = CMSModel.Pages.getPageById(item1['id']).toJSON()
            item2_json = CMSModel.Pages.getPageById(item2['id']).toJSON()
        except:
            raise amiweb.AppError("The host page can't be re-ordered.")

        PageTabsModel.swapPageOrder(ident, item1['id'], item2['id'])

        res = json.write({
            'item1_json': item1_json,
            'item2_json': item2_json})
        self._expireCache(ident)
        return res

    def _getHostPageJSON(self, ident):
        p_tab = PageTabsModel.getTabByIdent(ident)
        return json.write({'id': p_tab.ident, 'content': p_tab.host_page_name, 'name': p_tab.host_page_name, 'type': 'host_page', 'ident': ident})

    def _expireCache(self, ident):
        ns = PageTabsModel.getTabByIdent(ident)
        AmiCache.expireCache(ns.host_page_id)
        for p in PageTabsModel.getPagesByIdent(ident):
            AmiCache.expireCache(p.id)


class PageTabsRender:
    """
    Renders the actual PageTabs HTML
    """
    def render(self, ident, edit_mode, as_list=0):
        if as_list:
            as_list = 1
        else:
            as_list = 0

        btn_manage = html_helpers.createActionLink("Manage tabs", "static_plugin/pagetabs/manage.png", "return CMS_PageTabs.viewManage(this, '%s', %s)" % ('%s', as_list))

        btn_manage = btn_manage % ident

        current_pid = amiweb.session()['current_page_id']
        p = PageTabsModel.getTabByIdent(ident)

        if not p:
            PageTabsModel.add(ident, current_pid)
            p = PageTabsModel.getTabByIdent(ident)

        #Check that p's page is found, else set it to current
        host_page = CMSModel.Pages.getPageById(p.host_page_id)

        if host_page == None:
            PageTabsModel.updateHostPage(p.id, current_pid)
            p.host_page_id = current_pid

        ns = {
          'host_page': CMSModel.Pages.getPageById(p.host_page_id),
          'current_page': CMSModel.Pages.getPageById(current_pid),
          'pages': PageTabsModel.getPagesByIdent(ident),
          'ident': ident,
          'host_page_name': p.host_page_name,
          'edit_mode': edit_mode,
          'btn_manage': btn_manage
        }

        if as_list:
            return render("skeletonz/plugins/pagetabs/view/list_items.tmpl", ns)
        else:
            return render("skeletonz/plugins/pagetabs/view/items.tmpl", ns)


##
# DB Version control
#
class PageTab_Initial(AmiGration):

    def __init__(self):
        self.table_prefix = getConfig().TABLE_PREFIX

    def up(self):
        self.createTable("plugin_tabs",
          IntCol("id", length=11, auto_increment=True),
          StringCol("ident", not_null=True, length=100),
          StringCol("pages", not_null=True),
          PrimaryKey("id"),
          UniqueKey("ident", "ident"),
          utf8=True,
          ignore_if_created=True
        )

    def down(self):
        self.dropTable("plugin_tabs")

class PageTab_AddHostPage(AmiGration):

    def __init__(self):
        self.table_prefix = getConfig().TABLE_PREFIX

    def up(self):
        self.addColumn("plugin_tabs",
          IntCol("host_page_id", length=11)
        )

    def down(self):
        self.removeColumn("plugin_tabs", "host_page_id")

class PageTab_AddHostPageName(AmiGration):

    def __init__(self):
        self.table_prefix = getConfig().TABLE_PREFIX

    def up(self):
        self.addColumn("plugin_tabs",
          StringCol("host_page_name", length=100, default="Main")
        )

    def down(self):
        self.removeColumn("plugin_tabs", "host_page_name")


PageTabsModel = PageTabsModel()
PageTabsRender = PageTabsRender()
